from typing import Tuple
from torch.utils.data import DataLoader
import pytest
import numpy as np
import pnlp

from hcgf.dataloader.dataset import GlmMapStyleDataset
from hcgf.data_model import DataItem
from hcgf.dataloader.data_collector import GlmDataCollector


def test_dataset(glm_data_file, glm_tokenizer):
    data = pnlp.read_file_to_list_dict(glm_data_file)
    ds = GlmMapStyleDataset(data, glm_tokenizer)
    assert isinstance(ds[0], DataItem)
    for v in ds:
        assert isinstance(v, DataItem)
        break
    assert len(ds) == len(data)


@pytest.mark.parametrize("max_len,expected", [
    (5, 8),
    (128, 10),
])
def test_dataset_max_len(glm_tokenizer, max_len, expected):
    data = [
        {"prompt": "爱爱爱爱", "completion": "谁"}
    ]
    ds = GlmMapStyleDataset(data, glm_tokenizer, max_len)
    v = ds[0]
    # 爱爱爱爱 encode后长度为7，超过后变为5，加上completion的三个：空白符、谁、EOS，共8个token
    assert len(v.input_ids) == expected


arr_dtype = np.int64

input_ids = np.array(
    [
        [20005, 94874, 94874, 94874, 150001, 150004, 20005, 88443, 20002],
        [20005, 94874, 150001, 150004, 20005, 84480, 20002, 20002, 20002],
    ],
    dtype=arr_dtype
)

labels = np.array(
    [
        [-100, -100, -100, -100, -100, 150004, 20005, 88443, 20002],
        [-100, -100, -100, 150004, 20005, 84480, 20002, -100, -100],
    ],
    dtype=arr_dtype
)

position_ids = np.array(
    [
        [[0, 1, 2, 3, 4, 5, 6, 7, 8],
         [0, 0, 0, 0, 0, 1, 2, 3, 4]],
        [[0, 1, 2, 3, 4, 5, 6, 7, 8],
         [0, 0, 0, 1, 2, 3, 4, 5, 6]],
    ],
    dtype=arr_dtype
)

attention_mask = np.array(
    [
        [[[False, False, False, False, False, True, True, True, True],
          [False, False, False, False, False, True, True, True, True],
          [False, False, False, False, False, True, True, True, True],
          [False, False, False, False, False, True, True, True, True],
          [False, False, False, False, False, True, True, True, True],
          [False, False, False, False, False, False, True, True, True],
          [False, False, False, False, False, False, False, True, True],
          [False, False, False, False, False, False, False, False, True],
          [False, False, False, False, False, False, False, False, False]]],
        [[[False, False, False, True, True, True, True, True, True],
          [False, False, False, True, True, True, True, True, True],
          [False, False, False, True, True, True, True, True, True],
          [False, False, False, False, True, True, True, True, True],
          [False, False, False, False, False, True, True, True, True],
          [False, False, False, False, False, False, True, True, True],
          [False, False, False, False, False, False, False, True, True],
          [False, False, False, False, False, False, False, False, True],
          [False, False, False, False, False, False, False, False, False]]],
    ],
    dtype=np.bool_
)


@pytest.mark.parametrize("inp_key,shape,expected", [
    ("input_ids", (2, 9), input_ids),
    ("position_ids", (2, 2, 9), position_ids),
    ("attention_mask", (2, 1, 9, 9), attention_mask),
    ("labels", (2, 9), labels),
])
def test_data_collector(mocked_dataset, inp_key, shape, expected):
    binp = GlmDataCollector.collate_fn(mocked_dataset)
    assert type(binp) == dict
    val = binp[inp_key]
    assert tuple(val.shape) == shape
    assert np.alltrue(val.numpy() == expected)


@pytest.mark.parametrize("func,expected", [
    ("train_dev_split", Tuple),
    ("load", DataLoader),
])
@pytest.mark.parametrize("batch_size", [1, 2, 3])
def test_data_loader(glm_dataloader, func, expected, batch_size):
    out = getattr(glm_dataloader, func)(batch_size=batch_size)
    assert isinstance(out, expected)
    if isinstance(out, DataLoader):
        dataloader = out
        data_len = len(glm_dataloader)
        num_batches = data_len // batch_size
        if data_len % batch_size != 0:
            num_batches += 1
        assert len(dataloader) == num_batches
    else:
        dataloader = out[0]
    for v in dataloader:
        input_ids = v["input_ids"]
        assert input_ids.shape[0] == batch_size
        break


@pytest.mark.parametrize("inp_key,shape", [
    ("input_ids", (1, 7)),
    ("position_ids", (1, 2, 7)),
    ("attention_mask", (1, 1, 7, 7)),
    ("labels", (1, 7)),
])
def test_dataset_collector(glm_tokenizer, inp_key, shape):
    data = [
        {"prompt": "你好", "completion": "谁"},
    ]
    ds = GlmMapStyleDataset(data, glm_tokenizer)
    binp = GlmDataCollector.collate_fn(ds)
    val = binp[inp_key]
    assert tuple(val.shape) == shape



import os
import shutil
import pytest

from hcgf.sft.lora_ft import GlmLora


@pytest.mark.slow
def test_8bit_mode(glm_data_file, glm_tune_param):
    params = glm_tune_param
    model_id = "THUDM/chatglm-6b"
    p1 = "./test_output1/"
    p2 = "./test_output2/"
    no_bnb = False
    try:
        import bitsandbytes as bnb
    except Exception:
        no_bnb = True
    if no_bnb:
        pass
    else:
        gl = GlmLora(model_id, load_in_8bit=True)
        params["out_dir"] = p1
        print("tuning...")
        (gl
        .load_data(glm_data_file, max_seq_len=32)
        .tune(**params))
        gl.eval()
        response, history = gl.chat("你是谁？")
        print(response)
        params["out_dir"] = p2
        print("\n\ntuning again...")
        gl.tune(**params)
        print("\n\ninference...")
        gl.load_pretrained(os.path.join(params["out_dir"], "ckpt/lora-ckpt-last-19.pt")).eval()
        response, history = gl.chat("你是谁？")
        print(response)
        assert 1, "should pass"
    
    for path in [p1, p2]:
        if os.path.exists(path):
            shutil.rmtree(path)


@pytest.mark.slow
def test_normal_mode(glm_data_file, glm_tune_param):
    params = glm_tune_param
    model_id = "THUDM/chatglm-6b"
    p1 = "./test_output1/"
    p2 = "./test_output2/"
    gl = GlmLora(model_id, device="cuda:0")
    params["out_dir"] = p1
    print("tuning...")
    (gl
    .load_data(glm_data_file, max_seq_len=32)
    .tune(**params))
    gl.eval()
    response, history = gl.chat("你是谁？")
    print(response)
    params["out_dir"] = p2
    print("\n\ntuning again...")
    gl.tune(**params)
    print("\n\ninference...")
    gl.load_pretrained(os.path.join(params["out_dir"], "ckpt/lora-ckpt-last-19.pt")).eval()
    response, history = gl.chat("你是谁？")
    print(response)
    assert 1, "should pass"
    for path in [p1, p2]:
        if os.path.exists(path):
            shutil.rmtree(path)


from typing import Dict, List

import torch
import torch.nn as nn
from transformers.tokenization_utils import PreTrainedTokenizer


from ..data_model import Tensor


def print_trainable_parameters(model: nn.Module) -> None:
    """
    Prints the number of trainable parameters in the model.
    """
    trainable_params = 0
    all_param = 0
    for _, param in model.named_parameters():
        all_param += param.numel()
        if param.requires_grad:
            trainable_params += param.numel()
    msg = f"trainable params: {trainable_params} || all params: {all_param} || "
    msg += f"trainable%: {100 * trainable_params / all_param}"
    print(msg)


def print_layer_info(model: nn.Module) -> None:
    for key, val in model.named_parameters():
        msg = "\t".join(map(str, (val.dtype, val.device, val.requires_grad, key)))
        print(msg)


def get_lora_state_dict(
        model: nn.Module, bias: str = "none") -> Dict[str, torch.Tensor]:
    """
    From https://github.com/microsoft/LoRA/
    """
    my_state_dict = model.state_dict()
    if bias == "none":
        return {k: my_state_dict[k] for k in my_state_dict if "lora_" in k}
    elif bias == "all":
        return {k: my_state_dict[k]
                for k in my_state_dict if "lora_" in k or "bias" in k}
    elif bias == "lora_only":
        to_return = {}
        for k in my_state_dict:
            if "lora_" in k:
                to_return[k] = my_state_dict[k]
                bias_name = k.split("lora_")[0] + "bias"
                if bias_name in my_state_dict:
                    to_return[bias_name] = my_state_dict[bias_name]
        return to_return
    else:
        raise NotImplementedError


def create_token_tensor_list(tokenizer: PreTrainedTokenizer, tokens: List[str]
    ) -> List[Tensor["L", torch.LongTensor]]:
    tensor_list = []
    for token in tokens:
        tids = tokenizer(
            token, add_special_tokens=False, return_tensors="pt"
        )["input_ids"].squeeze()[1:]
        tensor_list.append(tids)
    return tensor_list


import pytest

import torch

from hcgf.utils.utils import create_token_tensor_list


@pytest.mark.parametrize("tokens", [
    ["。"],
    ["。", "！"],
])
def test_create_token_tensor_list(glm_tokenizer, tokens):
    tensor_list = create_token_tensor_list(glm_tokenizer, tokens)
    assert type(tensor_list) == list
    assert len(tensor_list) == len(tokens)
    assert isinstance(tensor_list[0], torch.LongTensor)
    assert len(tensor_list[0]) == 1


"""Tokenization classes for ChatGLM."""
from typing import List, Optional, Union
import os

from transformers.tokenization_utils import PreTrainedTokenizer
from icetk.text_tokenizer import TextTokenizer
import icetk.sentencepiece_model_pb2 as sp_model
from transformers.utils import logging, PaddingStrategy
from transformers.tokenization_utils_base import EncodedInput, BatchEncoding
from typing import Dict
import numpy as np

logger = logging.get_logger(__name__)

PRETRAINED_POSITIONAL_EMBEDDINGS_SIZES = {
    "THUDM/chatglm-6b": 2048,
}


class SPTokenizer:
    def __init__(
        self,
        vocab_file,
        max_blank_length=80,
        byte_fallback=True,
    ):
        assert vocab_file is not None
        self.vocab_file = vocab_file
        self.special_tokens = ["[MASK]", "[gMASK]", "[sMASK]", "<unused_0>", "<sop>", "<eop>", "<ENC>", "<dBLOCK>"]
        self.max_blank_length = max_blank_length
        self.byte_fallback = byte_fallback
        self.text_tokenizer = self._build_text_tokenizer(encode_special_tokens=False)
        self.special_text_tokenizer = self._build_text_tokenizer(encode_special_tokens=True)

    @staticmethod
    def _configure_tokenizer(
        text_tokenizer: TextTokenizer,
        special_tokens: List[str],
        max_blank_length: int,
        byte_fallback: bool,
        encode_special_tokens=False,
    ):
        # special token
        special_token_type = 4 if encode_special_tokens else 3  # 3 - CONTROL, 4 - USER_DEFINE
        for token in special_tokens:
            text_tokenizer.proto.pieces.append(
                sp_model.ModelProto.SentencePiece(piece=token, score=0.0, type=special_token_type)
            )
        # whitespaces
        for token in [SPTokenizer.get_tab_token()] + [
            SPTokenizer.get_blank_token(i) for i in range(2, max_blank_length + 1)
        ]:
            text_tokenizer.proto.pieces.append(sp_model.ModelProto.SentencePiece(piece=token, score=0.0, type=4))
        # byte fallback
        if byte_fallback:
            text_tokenizer.proto.trainer_spec.byte_fallback = True
            for i in range(256):
                text_tokenizer.proto.pieces.append(
                    sp_model.ModelProto.SentencePiece(piece="<0x{:02X}>".format(i), score=0.0, type=6)
                )
        text_tokenizer.refresh()

    def _build_text_tokenizer(self, encode_special_tokens=False):
        tokenizer = TextTokenizer(self.vocab_file)
        self._configure_tokenizer(
            tokenizer, self.special_tokens, self.max_blank_length, self.byte_fallback, encode_special_tokens
        )
        return tokenizer

    def _get_text_tokenizer(self, encode_special_tokens=False):
        if encode_special_tokens:
            return self.special_text_tokenizer
        else:
            return self.text_tokenizer

    @staticmethod
    def get_blank_token(length: int):
        assert length >= 2
        return f"<|blank_{length}|>"

    @staticmethod
    def get_tab_token():
        return f"<|tab|>"

    @property
    def num_image_tokens(self):
        return 20000

    @property
    def num_text_tokens(self):
        return self.text_tokenizer.num_tokens

    @property
    def num_tokens(self):
        return self.num_image_tokens + self.num_text_tokens

    @staticmethod
    def _encode_whitespaces(text: str, max_len: int = 80):
        text = text.replace("\t", SPTokenizer.get_tab_token())
        for i in range(max_len, 1, -1):
            text = text.replace(" " * i, SPTokenizer.get_blank_token(i))
        return text

    def _preprocess(self, text: str, linebreak=True, whitespaces=True):
        if linebreak:
            text = text.replace("\n", "<n>")
        if whitespaces:
            text = self._encode_whitespaces(text, max_len=self.max_blank_length)
        return text

    def encode(
        self, text: str, linebreak=True, whitespaces=True, special_tokens=False, add_dummy_prefix=True
    ) -> List[int]:
        """
        @param text: Text to encode.
        @param linebreak: Whether to encode newline (\n) in text.
        @param whitespaces: Whether to encode multiple whitespaces or tab in text, useful for source code encoding.
        @param special_tokens: Whether to encode special token ([MASK], [gMASK], etc.) in text.
        @param add_dummy_prefix: Whether to add dummy blank space in the beginning.
        """
        text = self._preprocess(text, linebreak, whitespaces)
        if not add_dummy_prefix:
            text = "<n>" + text
        tmp = self._get_text_tokenizer(encode_special_tokens=special_tokens).encode(text)
        tokens = [x + self.num_image_tokens for x in tmp]
        return tokens if add_dummy_prefix else tokens[2:]

    def decode(self, text_ids: List[int], special_tokens=False) -> str:
        ids = [int(_id) - self.num_image_tokens for _id in text_ids]
        ids = [_id for _id in ids if _id >= 0]
        text = self._get_text_tokenizer(encode_special_tokens=special_tokens).decode(ids)
        text = text.replace("<n>", "\n")
        text = text.replace(SPTokenizer.get_tab_token(), "\t")
        for i in range(2, self.max_blank_length + 1):
            text = text.replace(self.get_blank_token(i), " " * i)
        return text

    def tokenize(
        self, text: str, linebreak=True, whitespaces=True, special_tokens=False, add_dummy_prefix=True
    ) -> List[str]:
        """
        @param text: Text to encode.
        @param linebreak: Whether to encode newline (\n) in text.
        @param whitespaces: Whether to encode multiple whitespaces or tab in text, useful for source code encoding.
        @param special_tokens: Whether to encode special token ([MASK], [gMASK], etc.) in text.
        @param add_dummy_prefix: Whether to add dummy blank space in the beginning.
        """
        text = self._preprocess(text, linebreak, whitespaces)
        if not add_dummy_prefix:
            text = "<n>" + text
        tokens = self._get_text_tokenizer(encode_special_tokens=special_tokens).tokenize(text)
        return tokens if add_dummy_prefix else tokens[2:]

    def __getitem__(self, x: Union[int, str]):
        if isinstance(x, int):
            if x < self.num_image_tokens:
                return "<image_{}>".format(x)
            else:
                return self.text_tokenizer.convert_id_to_token(x - self.num_image_tokens)
        elif isinstance(x, str):
            if x.startswith("<image_") and x.endswith(">") and x[7:-1].isdigit():
                return int(x[7:-1])
            else:
                return self.text_tokenizer.convert_token_to_id(x) + self.num_image_tokens
        else:
            raise ValueError("The key should be str or int.")


class ChatGLMTokenizer(PreTrainedTokenizer):
    """
    Construct a ChatGLM tokenizer. Based on byte-level Byte-Pair-Encoding.
    Args:
        vocab_file (`str`):
            Path to the vocabulary file.
    """

    vocab_files_names = {"vocab_file": "ice_text.model"}
    max_model_input_sizes = PRETRAINED_POSITIONAL_EMBEDDINGS_SIZES
    model_input_names = ["input_ids", "attention_mask", "position_ids"]

    def __init__(
            self,
            vocab_file,
            do_lower_case=False,
            remove_space=False,
            bos_token='sop',
            eos_token='eos',
            eop_token='eop',
            mask_token='[MASK]',
            gmask_token='[gMASK]',
            padding_side="left",
            **kwargs
    ) -> None:
        super().__init__(
            do_lower_case=do_lower_case,
            remove_space=remove_space,
            padding_side=padding_side,
            **kwargs
        )

        self.do_lower_case = do_lower_case
        self.remove_space = remove_space
        self.vocab_file = vocab_file

        self.bos_token = bos_token
        self.eos_token = eos_token
        self.eop_token = eop_token
        self.mask_token = mask_token
        self.gmask_token = gmask_token

        self.sp_tokenizer = SPTokenizer(vocab_file)

        """ Initialisation """

    @property
    def eop_token_id(self) -> Optional[int]:
        """
        `Optional[int]`: Id of the end of sentence token in the vocabulary. Returns `None` if the token has not been
        set.
        """
        if self.eop_token is None:
            return None
        return self.convert_tokens_to_ids(self.eop_token)

    @property
    def vocab_size(self):
        """ Returns vocab size """
        return self.sp_tokenizer.num_tokens

    def get_vocab(self):
        """ Returns vocab as a dict """
        vocab = {self._convert_id_to_token(i): i for i in range(self.vocab_size)}
        vocab.update(self.added_tokens_encoder)
        return vocab

    def preprocess_text(self, inputs):
        if self.remove_space:
            outputs = " ".join(inputs.strip().split())
        else:
            outputs = inputs

        if self.do_lower_case:
            outputs = outputs.lower()

        return outputs

    def _tokenize(self, text, **kwargs):
        """ Returns a tokenized string. """
        text = self.preprocess_text(text)

        seq = self.sp_tokenizer.tokenize(text)

        return seq

    def decode(
            self,
            token_ids: Union[List[int], List[List[int]]],
            skip_special_tokens: bool = False,
            clean_up_tokenization_spaces: bool = True,
            spaces_between_special_tokens: bool = True,
            **kwargs
    ) -> str:
        if isinstance(token_ids[0], list):
            tokens = []
            for single_token_ids in token_ids:
                if self.pad_token_id in single_token_ids:  # remove pad
                    single_token_ids = list(filter((self.pad_token_id).__ne__, single_token_ids))
                tokens.append(self.sp_tokenizer.decode(single_token_ids))
            return (tokens)
        else:
            if self.pad_token_id in token_ids:  # remove pad
                token_ids = list(filter((self.pad_token_id).__ne__, token_ids))
            return self.sp_tokenizer.decode(token_ids)

    def _convert_token_to_id(self, token):
        """ Converts a token (str) in an id using the vocab. """
        return self.sp_tokenizer[token]

    def _convert_id_to_token(self, index):
        """Converts an index (integer) in a token (str) using the vocab."""
        return self.sp_tokenizer[index]

    def save_vocabulary(self, save_directory, filename_prefix=None):
        """
        Save the vocabulary and special tokens file to a directory.
        Args:
            save_directory (`str`):
                The directory in which to save the vocabulary.
            filename_prefix (`str`, *optional*):
                An optional prefix to add to the named of the saved files.
        Returns:
            `Tuple(str)`: Paths to the files saved.
        """
        if os.path.isdir(save_directory):
            vocab_file = os.path.join(
                save_directory, self.vocab_files_names["vocab_file"]
            )
        else:
            vocab_file = save_directory

        with open(self.vocab_file, 'rb') as fin:
            proto_str = fin.read()

        with open(vocab_file, "wb") as writer:
            writer.write(proto_str)

        return (vocab_file,)

    def build_inputs_with_special_tokens(
            self, token_ids_0: List[int], token_ids_1: Optional[List[int]] = None
    ) -> List[int]:
        """
        Build model inputs from a sequence or a pair of sequence for sequence classification tasks by concatenating and
        adding special tokens. A BERT sequence has the following format:
        - single sequence: `[CLS] X [SEP]`
        - pair of sequences: `[CLS] A [SEP] B [SEP]`
        Args:
            token_ids_0 (`List[int]`):
                List of IDs to which the special tokens will be added.
            token_ids_1 (`List[int]`, *optional*):
                Optional second list of IDs for sequence pairs.
        Returns:
            `List[int]`: List of [input IDs](../glossary#input-ids) with the appropriate special tokens.
        """
        mask_ids = self.sp_tokenizer[self.mask_token]
        gmask_ids = self.sp_tokenizer[self.gmask_token]
        eop_id = self.sp_tokenizer[self.eop_token]
        if mask_ids not in token_ids_0 and gmask_ids not in token_ids_0:
            token_ids_0 += [gmask_ids]

        if token_ids_0[-1] != mask_ids and token_ids_0[-1] != gmask_ids:
            token_ids_0 += [self.sp_tokenizer[self.eos_token]]

        token_ids_0 += [self.sp_tokenizer[self.bos_token]]

        if token_ids_1 is not None:
            if not token_ids_1 or token_ids_1[-1] != eop_id:
                token_ids_1 += [eop_id]
            token_ids_0 += token_ids_1

        return token_ids_0

    def _pad(
        self,
        encoded_inputs: Union[Dict[str, EncodedInput], BatchEncoding],
        max_length: Optional[int] = None,
        padding_strategy: PaddingStrategy = PaddingStrategy.DO_NOT_PAD,
        pad_to_multiple_of: Optional[int] = None,
        return_attention_mask: Optional[bool] = None,
    ) -> dict:
        """
        Pad encoded inputs (on left/right and up to predefined length or max length in the batch)
        Args:
            encoded_inputs:
                Dictionary of tokenized inputs (`List[int]`) or batch of tokenized inputs (`List[List[int]]`).
            max_length: maximum length of the returned list and optionally padding length (see below).
                Will truncate by taking into account the special tokens.
            padding_strategy: PaddingStrategy to use for padding.
                - PaddingStrategy.LONGEST Pad to the longest sequence in the batch
                - PaddingStrategy.MAX_LENGTH: Pad to the max length (default)
                - PaddingStrategy.DO_NOT_PAD: Do not pad
                The tokenizer padding sides are defined in self.padding_side:
                    - 'left': pads on the left of the sequences
                    - 'right': pads on the right of the sequences
            pad_to_multiple_of: (optional) Integer if set will pad the sequence to a multiple of the provided value.
                This is especially useful to enable the use of Tensor Core on NVIDIA hardware with compute capability
                `>= 7.5` (Volta).
            return_attention_mask:
                (optional) Set to False to avoid returning attention mask (default: set to model specifics)
        """
        # Load from model defaults
        bos_token_id = self.sp_tokenizer[self.bos_token]
        mask_token_id = self.sp_tokenizer[self.mask_token]
        gmask_token_id = self.sp_tokenizer[self.gmask_token]
        assert self.padding_side == "left"

        required_input = encoded_inputs[self.model_input_names[0]]
        seq_length = len(required_input)

        if padding_strategy == PaddingStrategy.LONGEST:
            max_length = len(required_input)

        if max_length is not None and pad_to_multiple_of is not None and (max_length % pad_to_multiple_of != 0):
            max_length = ((max_length // pad_to_multiple_of) + 1) * pad_to_multiple_of

        needs_to_be_padded = padding_strategy != PaddingStrategy.DO_NOT_PAD and len(required_input) != max_length

        # Initialize attention mask if not present.
        if max_length is not None:
            if "attention_mask" not in encoded_inputs:
                if bos_token_id in required_input:
                    context_length = required_input.index(bos_token_id)
                else:
                    context_length = seq_length
                attention_mask = np.ones((1, seq_length, seq_length))
                attention_mask = np.tril(attention_mask)
                attention_mask[:, :, :context_length] = 1
                attention_mask = np.bool_(attention_mask < 0.5)
                encoded_inputs["attention_mask"] = attention_mask

            if "position_ids" not in encoded_inputs:
                position_ids = np.arange(seq_length, dtype=np.int64)
                mask_token = mask_token_id if mask_token_id in required_input else gmask_token_id
                if mask_token in required_input:
                    mask_position = required_input.index(mask_token)
                    position_ids[context_length:] = mask_position
                block_position_ids = np.concatenate(
                    [np.zeros(context_length, dtype=np.int64),
                     np.arange(1, seq_length - context_length + 1, dtype=np.int64)])
                encoded_inputs["position_ids"] = np.stack([position_ids, block_position_ids], axis=0)

        if needs_to_be_padded:
            difference = max_length - len(required_input)

            if "attention_mask" in encoded_inputs:
                encoded_inputs["attention_mask"] = np.pad(encoded_inputs["attention_mask"],
                                                          pad_width=[(0, 0), (difference, 0), (difference, 0)],
                                                          mode='constant', constant_values=True)
            if "token_type_ids" in encoded_inputs:
                encoded_inputs["token_type_ids"] = [self.pad_token_type_id] * difference + encoded_inputs[
                    "token_type_ids"
                ]
            if "special_tokens_mask" in encoded_inputs:
                encoded_inputs["special_tokens_mask"] = [1] * difference + encoded_inputs["special_tokens_mask"]
            if "position_ids" in encoded_inputs:
                encoded_inputs["position_ids"] = np.pad(encoded_inputs["position_ids"],
                                                        pad_width=[(0, 0), (difference, 0)])
            encoded_inputs[self.model_input_names[0]] = [self.pad_token_id] * difference + required_input

        return encoded_inputs



from typing import Optional, List, Tuple
import copy

import torch
import torch.nn as nn
from transformers.modeling_utils import PreTrainedModel
from transformers.generation.utils import (
    LogitsProcessorList, 
    StoppingCriteriaList, 
    StoppingCriteria
)
from peft import get_peft_model, LoraConfig, TaskType

from ..utils import print_trainable_parameters, create_token_tensor_list
from ..dataloader import GlmDataLoader
from ..data_model import Tensor
from ..trainer import Trainer


from .chatglm import (
    ChatGLMForConditionalGeneration, 
    ChatGLMTokenizer, 
    InvalidScoreLogitsProcessor
)


class CustomStoppingCriteria(StoppingCriteria):

    def __init__(
        self, 
        stop_tensor_list: List[Tensor["L", torch.LongTensor]], 
        device: torch.device, 
        encounters: int = 1
    ):
        super().__init__()
        self.stops = [stop.to(device) for stop in stop_tensor_list]
        self.encounters = encounters

    def __call__(
        self, 
        input_ids: Tensor["B,L", torch.LongTensor], 
        scores: torch.FloatTensor
    ):
        count = 0
        for stop in self.stops:
            if torch.all((stop == input_ids[0][-len(stop):])).item():
                count += 1
            if count >= self.encounters:
                return True
        return False


class CastOutputToFloat(nn.Sequential):
    def forward(self, x):
        return super().forward(x).to(torch.float32)


class GlmLora:

    def __init__(
        self,
        model_id: str,
        device: Optional[str] = None,
        lora_r: int = 8,
        lora_alpha: int = 32,
        lora_dropout: float = 0.1,
        load_in_8bit: bool = False,
    ):
        print(f"Loading tokenizer and model of {model_id}")
        self.load_in_8bit = load_in_8bit
        if self.load_in_8bit:
            import bitsandbytes as bnb
            self.device = None
            model = self._load_8bit_glm(model_id)
        else:
            self.device = device or "cuda"
            model = self._load_glm(model_id)
        self.tokenizer = ChatGLMTokenizer.from_pretrained(model_id)
        self.config = LoraConfig(
            # peft config
            peft_type="LORA",
            task_type=TaskType.CAUSAL_LM,
            inference_mode=False,
            # lora config
            target_modules=["query_key_value"],
            r=lora_r,
            lora_alpha=lora_alpha,
            lora_dropout=lora_dropout,
            enable_lora=[True, False, True],
            bias="none",
        )
        print("Processing peft model")
        self.model = get_peft_model(model, self.config)
        print_trainable_parameters(self.model)

        self.stop_tokens = ["问题："]
        self.builtin_stop_tensor_list = self.init_stop_tensor_list()

    def __cast_small_to(self, dtype: torch.Type):
        for name, param in self.model.named_parameters():
            if param.ndim == 1 and param.dtype != dtype:
                param.data = param.data.to(dtype)
    
    def __cast_lora_to(self, dtype: torch.Type):
        for name, param in self.model.named_parameters():
            if "lora_" in name:
                param.data = param.data.to(dtype)

    def _load_8bit_glm(self, model_id: str) -> PreTrainedModel:
        model = ChatGLMForConditionalGeneration.from_pretrained(
            model_id, load_in_8bit=True, device_map="auto")
        for param in model.parameters():
            param.requires_grad = False
        return model

    def _load_glm(self, model_id: str) -> PreTrainedModel:
        model = ChatGLMForConditionalGeneration.from_pretrained(
            model_id).to(self.device)
        for param in model.parameters():
            param.requires_grad = False
        return model

    def load_pretrained(self, pt_path: str):
        static = torch.load(pt_path)
        self.model.load_state_dict(static, strict=False)
        return self

    def load_data(self, data_path: str, max_seq_len: int = 512):
        print(f"Loading data with max_seq_len: {max_seq_len}")
        self.dataloader = GlmDataLoader(data_path, self.tokenizer, max_seq_len)
        return self

    def tune(
        self,
        batch_size: int = 1,
        lr: float = 2e-4,
        num_epochs: int = 10,
        warmup_steps: Optional[int] = None,
        accumulate_steps: Optional[int] = 32,
        out_dir: str = "./output/",
        print_every: int = 10,
    ):
        """
        Note
        -----------
        if `warmup_steps` is None, will use one epoch to warmup by default
        if `accumulate_steps` is None, will use 1 as default
        """
        print("Switch to training mode...")
        if self.load_in_8bit:
            self.__cast_small_to(torch.float32)
        self.model.lm_head = CastOutputToFloat(self.model.lm_head)
        self.__cast_lora_to(torch.float32)
        # turn on when infer
        self.model.config.use_cache = False
        if self.device is not None:
            self.model.to(self.device).train()
        else:
            self.model.train()
        trainer = Trainer(
            lr,
            num_epochs,
            warmup_steps,
            accumulate_steps,
            out_dir,
            device=self.device,
            print_every=print_every,
        )
        train_dl, dev_dl = self.dataloader.train_dev_split(batch_size)
        print("Begining tunning")
        trainer.train(self.model, train_dl, dev_dl)

    def eval(self, quant_bit: Optional[int] = None):
        print("Switch to inference mode...")
        self.model.half()
        if self.load_in_8bit:
            self.__cast_lora_to(torch.float32)
        self.model.base_model.peft_config.inference_mode = True
        self.model.config.use_cache = True
        if quant_bit is not None:
            self.model.quantize(quant_bit)
        # 8bit do not use device, device is None
        if self.device is not None:
            self.model.to(self.device).eval()
        else:
            self.model.eval()
    
    @torch.no_grad()
    def stream_chat(
        self, 
        query: str, 
        history: List[Tuple[str, str]] = None, 
        max_length: int = 2048,
        do_sample=True, 
        top_p=0.7, 
        temperature=0.95, 
        logits_processor=None, 
        stopping_criteria=None,
        **kwargs
    ):
        "From ChatGLM Model"
        if logits_processor is None:
            logits_processor = LogitsProcessorList()
        logits_processor.append(InvalidScoreLogitsProcessor())
        if stopping_criteria is None:
            stopping_criteria = StoppingCriteriaList()
        gen_kwargs = {
            "max_length": max_length, 
            "do_sample": do_sample, 
            "top_p": top_p,
            "temperature": temperature, 
            "logits_processor": logits_processor,
            "stopping_criteria": stopping_criteria,
            **kwargs
        }
        if not history:
            prompt = query
        else:
            prompt = ""
            for i, (old_query, response) in enumerate(history):
                prompt += "[Round {}]\n 问：{}\n 答：{}\n".format(i, old_query, response)
            prompt += "[Round {}]\n 问：{}\n 答：".format(len(history), query)
        inputs = self.tokenizer([prompt], return_tensors="pt")
        inputs = inputs.to(self.model.device)
        for outputs in self.model.stream_generate(
            inputs["input_ids"], **gen_kwargs
        ):
            outputs = outputs.tolist()[0][len(inputs["input_ids"][0]):]
            response = self.tokenizer.decode(outputs)
            response = self.model.process_response(response)
            new_history = history + [(query, response)]
            yield response, new_history

    def chat(
        self, 
        inp: str, 
        history: List[Tuple[str, str]] = None, 
        max_len: int = 512, 
        stop: List[str] = []
    ):
        if not history:
            history = []
        
        if stop:
            stop_tokens = [v for v in stop if v not in self.stop_tokens]
            custom_stop_tensor_list = create_token_tensor_list(self.tokenizer, stop_tokens)
            stop_tensor_list = self.builtin_stop_tensor_list + custom_stop_tensor_list
        else:
            stop_tensor_list = self.builtin_stop_tensor_list

        custom_stop_list = [CustomStoppingCriteria(stop_tensor_list, self.model.device)]
        
        for response, history in self.stream_chat(
            inp, history, max_len,
            stopping_criteria=StoppingCriteriaList(custom_stop_list),
        ):
            ...
        return response, history
    
    def init_stop_tensor_list(self) -> List[Tensor["L", torch.LongTensor]]:
        init_tensor_list = [
            torch.LongTensor([20002]),  # eos
            torch.LongTensor([150005]), # eop
        ]
        new_add = create_token_tensor_list(self.tokenizer, self.stop_tokens)
        return init_tensor_list + new_add


from dataclasses import dataclass
import pytest

import torch
import torch.nn as nn

from hcgf.trainer.trainer import Trainer


@dataclass
class Output:
    loss: torch.FloatTensor


class Model(nn.Module):
    def __init__(self):
        super().__init__()
        self.mlp = nn.Linear(10, 2)
        self.dropout = nn.Dropout(0.1)
        self.softmax = nn.LogSoftmax(dim=1)

    def forward(self, input_ids, labels):
        z = self.mlp(input_ids)
        z = self.dropout(z)
        z = self.softmax(z)
        loss = loss_fn(z, labels)
        return Output(loss=loss)


loss_fn = nn.NLLLoss()


def gen_inp():
    return torch.randn(2, 10)


def gen_oup():
    return torch.randint(0, 2, (2,))


train_loader = [
    {
        "input_ids": gen_inp(),
        "labels": gen_oup(),
    } for i in range(20)
]

model = Model()


@pytest.mark.parametrize("lr", [1e-1])
@pytest.mark.parametrize("num_epochs", [1, 2])
@pytest.mark.parametrize("warmup_steps", [None, 10])
@pytest.mark.parametrize("accumulate_steps", [1, 8])
def test_trainer(lr, num_epochs, warmup_steps, accumulate_steps):
    trainer = Trainer(
        lr,
        num_epochs,
        warmup_steps,
        accumulate_steps,
        "test_output/",
        "cpu",
        10,
    )
    trainer.train(model, train_loader, train_loader)



from torch.nn import Linear
from torch.nn.parameter import Parameter

import bz2
import torch
import base64
import ctypes
from transformers.utils import logging

from typing import List
from functools import partial

logger = logging.get_logger(__name__)

try:
    from cpm_kernels.kernels.base import LazyKernelCModule, KernelFunction, round_up

    class Kernel:
        def __init__(self, code: bytes, function_names: List[str]):
            self.code = code
            self._function_names = function_names
            self._cmodule = LazyKernelCModule(self.code)

            for name in self._function_names:
                setattr(self, name, KernelFunction(self._cmodule, name))

    quantization_code = "$ "

    kernels = Kernel(
        bz2.decompress(base64.b64decode(quantization_code)),
        [
            "int4WeightCompression",
            "int4WeightExtractionFloat",
            "int4WeightExtractionHalf",
            "int8WeightExtractionFloat",
            "int8WeightExtractionHalf",
        ],
    )
except Exception as exception:
    kernels = None
    logger.warning("Failed to load cpm_kernels:" + str(exception))


class W8A16Linear(torch.autograd.Function):
    @staticmethod
    def forward(ctx, inp: torch.Tensor, quant_w: torch.Tensor, scale_w: torch.Tensor, weight_bit_width):
        ctx.inp_shape = inp.size()
        ctx.weight_bit_width = weight_bit_width
        out_features = quant_w.size(0)
        inp = inp.contiguous().view(-1, inp.size(-1))
        weight = extract_weight_to_half(quant_w, scale_w, weight_bit_width)
        ctx.weight_shape = weight.size()
        output = inp.mm(weight.t())
        ctx.save_for_backward(inp, quant_w, scale_w)
        return output.view(*(ctx.inp_shape[:-1] + (out_features,)))

    @staticmethod
    def backward(ctx, grad_output: torch.Tensor):
        inp, quant_w, scale_w = ctx.saved_tensors
        weight = extract_weight_to_half(quant_w, scale_w, ctx.weight_bit_width)
        grad_output = grad_output.contiguous().view(-1, weight.size(0))
        grad_input = grad_output.mm(weight)
        grad_weight = grad_output.t().mm(inp)
        return grad_input.view(ctx.inp_shape), grad_weight.view(ctx.weight_shape), None, None


def compress_int4_weight(weight: torch.Tensor):  # (n, m)
    with torch.cuda.device(weight.device):
        n, m = weight.size(0), weight.size(1)
        assert m % 2 == 0
        m = m // 2
        out = torch.empty(n, m, dtype=torch.int8, device="cuda")
        stream = torch.cuda.current_stream()

        gridDim = (n, 1, 1)
        blockDim = (min(round_up(m, 32), 1024), 1, 1)

        kernels.int4WeightCompression(
            gridDim,
            blockDim,
            0,
            stream,
            [ctypes.c_void_p(weight.data_ptr()), ctypes.c_void_p(out.data_ptr()), ctypes.c_int32(n), ctypes.c_int32(m)],
        )
        return out


def extract_weight_to_half(weight: torch.Tensor, scale_list: torch.Tensor, source_bit_width: int):
    if source_bit_width == 8:
        func = kernels.int8WeightExtractionHalf
    elif source_bit_width == 4:
        func = kernels.int4WeightExtractionHalf
    else:
        assert False, "Unsupported bit-width"

    with torch.cuda.device(weight.device):
        n, m = weight.size(0), weight.size(1)
        out = torch.empty(n, m * (8 // source_bit_width), dtype=torch.half, device="cuda")
        stream = torch.cuda.current_stream()

        gridDim = (n, 1, 1)
        blockDim = (min(round_up(m, 32), 1024), 1, 1)

        func(
            gridDim,
            blockDim,
            0,
            stream,
            [
                ctypes.c_void_p(weight.data_ptr()),
                ctypes.c_void_p(scale_list.data_ptr()),
                ctypes.c_void_p(out.data_ptr()),
                ctypes.c_int32(n),
                ctypes.c_int32(m),
            ],
        )
        return out


class QuantizedLinear(Linear):
    def __init__(self, weight_bit_width: int, weight_tensor=None, bias_tensor=None, empty_init=False, *args, **kwargs):
        super(QuantizedLinear, self).__init__(*args, **kwargs)
        self.weight_bit_width = weight_bit_width

        shape = self.weight.shape
        del self.weight

        if weight_tensor is None or empty_init:
            self.weight = torch.empty(
                shape[0], shape[1] * weight_bit_width // 8, dtype=torch.int8, device=kwargs["device"]
            )
            self.weight_scale = torch.empty(shape[0], dtype=kwargs["dtype"], device=kwargs["device"])
        else:
            self.weight_scale = (weight_tensor.abs().max(dim=-1).values / ((2 ** (weight_bit_width - 1)) - 1)).half()
            self.weight = torch.round(weight_tensor / self.weight_scale[:, None]).to(torch.int8)
            if weight_bit_width == 4:
                self.weight = compress_int4_weight(self.weight)

        self.weight = Parameter(self.weight.to(kwargs["device"]), requires_grad=False)
        self.weight_scale = Parameter(self.weight_scale.to(kwargs["device"]), requires_grad=False)
        if bias_tensor is not None:
            self.bias = Parameter(bias_tensor.to(kwargs["device"]), requires_grad=False)
        else:
            self.bias = None

    def forward(self, input):
        output = W8A16Linear.apply(input, self.weight, self.weight_scale, self.weight_bit_width)
        if self.bias is not None:
            output = output + self.bias
        return output


def quantize(model, weight_bit_width, empty_init=False, **kwargs):
    """Replace fp16 linear with quantized linear"""

    for layer in model.layers:
        layer.attention.query_key_value = QuantizedLinear(
            weight_bit_width=weight_bit_width,
            weight_tensor=layer.attention.query_key_value.weight.to(torch.cuda.current_device()),
            bias_tensor=layer.attention.query_key_value.bias,
            in_features=layer.attention.query_key_value.in_features,
            out_features=layer.attention.query_key_value.out_features,
            bias=True,
            dtype=torch.half,
            device=layer.attention.query_key_value.weight.device,
            empty_init=empty_init
        )
        layer.attention.dense = QuantizedLinear(
            weight_bit_width=weight_bit_width,
            weight_tensor=layer.attention.dense.weight.to(torch.cuda.current_device()),
            bias_tensor=layer.attention.dense.bias,
            in_features=layer.attention.dense.in_features,
            out_features=layer.attention.dense.out_features,
            bias=True,
            dtype=torch.half,
            device=layer.attention.dense.weight.device,
            empty_init=empty_init
        )
        layer.mlp.dense_h_to_4h = QuantizedLinear(
            weight_bit_width=weight_bit_width,
            weight_tensor=layer.mlp.dense_h_to_4h.weight.to(torch.cuda.current_device()),
            bias_tensor=layer.mlp.dense_h_to_4h.bias,
            in_features=layer.mlp.dense_h_to_4h.in_features,
            out_features=layer.mlp.dense_h_to_4h.out_features,
            bias=True,
            dtype=torch.half,
            device=layer.mlp.dense_h_to_4h.weight.device,
            empty_init=empty_init
        )
        layer.mlp.dense_4h_to_h = QuantizedLinear(
            weight_bit_width=weight_bit_width,
            weight_tensor=layer.mlp.dense_4h_to_h.weight.to(torch.cuda.current_device()),
            bias_tensor=layer.mlp.dense_4h_to_h.bias,
            in_features=layer.mlp.dense_4h_to_h.in_features,
            out_features=layer.mlp.dense_4h_to_h.out_features,
            bias=True,
            dtype=torch.half,
            device=layer.mlp.dense_4h_to_h.weight.device,
            empty_init=empty_init
        )
    return model


import time
import os
from typing import Optional

import torch
import torch.nn as nn
from torch.utils.data import DataLoader
from transformers import get_linear_schedule_with_warmup
import pnlp

from ..utils import get_lora_state_dict


class Trainer:

    def __init__(
        self,
        lr: float,
        num_epochs: int,
        warmup_steps: int,
        accumulate_steps: int,
        out_dir: str,
        device: Optional[str],
        print_every: int,
    ):
        self.lr = lr
        self.num_epochs = num_epochs
        self.warmup_steps = warmup_steps
        self.accumulate_steps = accumulate_steps
        self.out_dir = out_dir
        self.device = device
        self.print_every = max(print_every, 1)

    def train(
        self,
        model: nn.Module,
        train_dataloader: DataLoader,
        dev_dataloader: DataLoader
    ) -> None:
        ckpt_path = os.path.join(self.out_dir, "ckpt")
        pnlp.check_dir(self.out_dir, ckpt_path)
        start_time = time.perf_counter()

        optimizer = torch.optim.AdamW(model.parameters(), lr=self.lr)

        if self.accumulate_steps is None:
            accumulate_steps = 1
        else:
            accumulate_steps = self.accumulate_steps
        batch_num = len(train_dataloader)
        epoch_update_steps = int(batch_num / accumulate_steps)
        if self.warmup_steps is None:
            warmup_update_steps = epoch_update_steps
        else:
            warmup_update_steps = self.warmup_steps
        total_update_steps = epoch_update_steps * self.num_epochs

        # stops when 3 epochs do not improve
        early_stop_steps = 3 * batch_num
        # every 1/3 epoch do evaluation
        valid_steps = int(batch_num / 3)

        msg = f"Total data batches: {batch_num}, "
        msg += f"Epoch update steps: {epoch_update_steps}, "
        msg += f"Total update steps: {total_update_steps}, "
        msg += f"Warmup update steps: {warmup_update_steps}, "
        msg += f"Validation steps: {valid_steps}, "
        msg += f"Early stop steps: {early_stop_steps}"
        print(msg)

        lr_scheduler = get_linear_schedule_with_warmup(
            optimizer=optimizer,
            num_warmup_steps=warmup_update_steps,
            num_training_steps=total_update_steps,
        )

        val_best_loss = float("inf")
        total_step = 0
        last_improve = 0
        flag = False
        for epoch in range(self.num_epochs):
            train_loss = 0
            print(f"\n\nEpoch {epoch}/{self.num_epochs}")
            for step, batch in enumerate(train_dataloader, start=1):
                if self.device is not None:
                    batch = {k: v.to(self.device) for k, v in batch.items()}
                    dtype = torch.float16
                else:
                    dtype = torch.bfloat16
                # mix precision
                model.train()
                with torch.cuda.amp.autocast(dtype=dtype):
                    output = model(**batch)
                loss_b = output.loss.detach().float()
                train_loss += loss_b
                loss = output.loss / accumulate_steps
                loss.backward()
                total_step += 1

                if step % accumulate_steps == 0:
                    optimizer.step()
                    lr_scheduler.step()
                    optimizer.zero_grad()

                if step % self.print_every == 0:
                    msg = f"Step: {step}, "
                    msg += f"Loss: {train_loss/step:.4f}, "
                    msg += f"LearningRate: {lr_scheduler.get_last_lr()[0]:.6f} "
                    print(msg)

                if total_step % valid_steps == 0 and valid_steps < batch_num:
                    model.eval()
                    val_loss = self.eval(model, dev_dataloader)
                    if val_loss < val_best_loss:
                        val_best_loss = val_loss
                        last_improve = total_step
                        out_file_name = f"lora-ckpt-best-{total_step}.pt"
                        out_file_path = os.path.join(ckpt_path, out_file_name)
                        torch.save(get_lora_state_dict(model), out_file_path)

                    msg = "- " * 30 + "\n"
                    msg += f"Step(Batch)/Epoch: {step}/{epoch},  Total Step: {total_step},  "
                    msg += f"LearningRate: {lr_scheduler.get_last_lr()[0]:.6f}, \n"
                    msg += f"TrainLoss: {train_loss/step:.4f},  "
                    msg += f"ValidLoss: {val_loss:.4f}\n"
                    msg += "= " * 30
                    print(msg)

                if total_step - last_improve > early_stop_steps:
                    print("Early stop for no improvements...")
                    flag = True

            secs = int(time.perf_counter() - start_time)
            mins = secs / 60
            secs = secs % 60
            model.eval()
            val_loss = self.eval(model, dev_dataloader)
            msg = f"Epoch: {epoch} | time in {mins:.1f} minutes, {secs} seconds \n"
            msg += f"\tEpoch TrainLoss: {train_loss/batch_num:.4f}  \n"
            msg += f"\tEpoch ValidLoss: {val_loss:.4f}  "
            print(msg)

            if flag:
                break

        out_file_name = f"lora-ckpt-last-{total_step}.pt"
        out_file_path = os.path.join(ckpt_path, out_file_name)
        torch.save(get_lora_state_dict(model), out_file_path)

    def eval(self, model: nn.Module, dataloader: DataLoader) -> float:
        total_loss = 0
        steps = 0
        for step, batch in enumerate(dataloader, start=1):
            steps += 1
            if self.device is not None:
                batch = {k: v.to(self.device) for k, v in batch.items()}
                dtype = torch.float16
            else:
                dtype = torch.bfloat16
            with torch.cuda.amp.autocast(dtype=dtype):
                with torch.no_grad():
                    output = model(**batch)
            loss_b = output.loss
            total_loss += loss_b
        return total_loss / steps



"""
Referenced from `trl`
"""

import torch.nn as nn

from transfromers import AutoModel

from ..data_model import Hidden


class ValueHead(nn.Module):

    def __init__(self, config, **kwargs):
        super().__init__()
        self.dropout = nn.Dropout(0.1)
        hidden_size = config.hidden_size
        self.mlp = nn.Linear(hidden_size, 1, bias=False)

    def forward(self, hidden_states: Hidden) -> Hidden:
        # BxLxH
        output = self.dropout(hidden_states)
        # BxLx1
        output = self.mlp(output)
        return output


class RewardModel(nn.Module):

    def __init__(self, model_id: str):
        super().__init__()
        self.pretrained_model = AutoModel.from_pretrained(
            model_id)
        self.value_head = ValueHead(self.pretrained_model.config)

    def forward(
        self,
        input_ids=None,
        past_key_values=None,
        attention_mask=None,
        **kwargs
    ) -> Hidden:

        base_model_output = self.pretrained_model(
            input_ids=input_ids,
            past_key_values=past_key_values,
            attention_mask=attention_mask,
            output_hidden_states=True,
            **kwargs,
        )
        # BxLxH
        last_hidden_state = base_model_output.hidden_states[-1]
        # BxLx1 => BxL
        value = self.value_head(last_hidden_state).squeeze(-1)
        return value
